/*
 * Copyright (C) 2016 Socionext Inc.
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

#include <common.h>
#include <linux/io.h>
#include <linux/kernel.h>

#include "../sc64-regs.h"
#include "../sg-regs.h"
#include "local-init.h"

/* Pin Drive Control */
#define SG_DRVCTRL_BASE			(SG_CTRL_BASE | 0x1800)
#define SG_DRV2CTRL_BASE		(SG_CTRL_BASE | 0x1900)
#define SG_DRV3CTRL_BASE		(SG_CTRL_BASE | 0x1980)

static void sg_set_drvctrl_common(unsigned long reg_base, unsigned pin,
				  unsigned muxval, unsigned mux_bits,
				  unsigned shift_bits,
				  unsigned reg_stride)
{
	unsigned shift = pin * shift_bits % 32;
	unsigned long reg = reg_base + pin * shift_bits / 32 * reg_stride;
	u32 mask = (1U << mux_bits) - 1;
	u32 tmp;

	tmp = readl(reg);
	tmp &= ~(mask << shift);
	tmp |= (mask & muxval) << shift;
	writel(tmp, reg);
}

/* unused */
static void __maybe_unused sg_set_drvctrl(unsigned pin, unsigned muxval)
{
	sg_set_drvctrl_common(SG_DRVCTRL_BASE, pin, muxval, 1, 1, 4);
}

static void __maybe_unused sg_set_drv2ctrl(unsigned pin, unsigned muxval)
{
	sg_set_drvctrl_common(SG_DRV2CTRL_BASE, pin, muxval, 2, 2, 4);
}

static void __maybe_unused sg_set_drv3ctrl(unsigned pin, unsigned muxval)
{
	sg_set_drvctrl_common(SG_DRV3CTRL_BASE, pin, muxval, 3, 4, 4);
}

struct clkrst_data {
	unsigned long address;
	u32 data;
};

static const struct clkrst_data reset_data[] = {
	{ SC_RSTCTRL,  0x00000001 },
	{ SC_RSTCTRL3, 0x000001f7 },
	{ SC_RSTCTRL4, 0x0000f3ff },
	{ SC_RSTCTRL5, 0x0000007f },
	{ SC_RSTCTRL6, 0x0000010f },
	{ SC_RSTCTRL7, 0x00010707 }
};

static const struct clkrst_data clock_data[] = {
	{ SC_CLKCTRL,  0x00000001 },
	{ SC_CLKCTRL3, 0x00000137 },
	{ SC_CLKCTRL4, 0x0000f3ff },
	{ SC_CLKCTRL5, 0x0000007f },
	{ SC_CLKCTRL6, 0x0000010f },
	{ SC_CLKCTRL7, 0x00010007 }
};

/* local init code for LD20 */
void uniphier_ld20_local_init(void)
{
	int i;

	writel(0x00000000, 0x5F801014);	/* PINCTRL05 */
	writel(0x00000000, 0x5F801018);	/* PINCTRL06 */
	writel(0x00000000, 0x5F80101C);	/* PINCTRL07 */
	writel(0x00000000, 0x5F801020);	/* PINCTRL08 */
	writel(0x00000000, 0x5F801024);	/* PINCTRL09 */
	writel(0x00000000, 0x5F801028);	/* PINCTRL10 */
	writel(0x00000000, 0x5F80102C);	/* PINCTRL11 */
	writel(0x00000000, 0x5F801030);	/* PINCTRL12 */
	writel(0x00000F0F, 0x5F801034);	/* PINCTRL13 */
	writel(0x00000F0F, 0x5F801038);	/* PINCTRL14 */
	writel(0x00000000, 0x5F80103C);	/* PINCTRL15 */
	writel(0x00000000, 0x5F801040);	/* PINCTRL16 */
	writel(0x02020200, 0x5F801044);	/* PINCTRL17 */
	writel(0x0F0F0202, 0x5F801048);	/* PINCTRL18 */
	writel(0x0F0F0F0F, 0x5F80104C);	/* PINCTRL19 */
	writel(0x0F0F0F0F, 0x5F801050);	/* PINCTRL20 */
	writel(0x0F0F0F0F, 0x5F801054);	/* PINCTRL21 */
	writel(0x0F0F0F0F, 0x5F801058);	/* PINCTRL22 */
	writel(0x0F0F0F0F, 0x5F80105C);	/* PINCTRL23 */
	writel(0x0F0F0F0F, 0x5F801060);	/* PINCTRL24 */
	writel(0x0F0F0F0F, 0x5F801064);	/* PINCTRL25 */
	writel(0x0F0F0F0F, 0x5F801068);	/* PINCTRL26 */
	writel(0x0F0F0F0F, 0x5F80106C);	/* PINCTRL27 */
	writel(0x0F0F0F0F, 0x5F801070);	/* PINCTRL28 */
	writel(0x0F0F0F0F, 0x5F801074);	/* PINCTRL29 */
	writel(0x0F0F0F0F, 0x5F801078);	/* PINCTRL30 */
	writel(0x03030303, 0x5F80107C);	/* PINCTRL31 */
	writel(0x0303030F, 0x5F801080);	/* PINCTRL32 */
	writel(0x000F0F03, 0x5F801084);	/* PINCTRL33 */
	writel(0x00000000, 0x5F801088);	/* PINCTRL34 */
	writel(0x000F0F00, 0x5F80108C);	/* PINCTRL35 */
	writel(0x00000000, 0x5F801090);	/* PINCTRL36 */
	writel(0x0F0F0E0F, 0x5F801094);	/* PINCTRL37 */
	writel(0x0F0E0E0E, 0x5F801098);	/* PINCTRL38 */
	writel(0x0E0F0F0F, 0x5F80109C);	/* PINCTRL39 */
	writel(0x0F0F0E0F, 0x5F8010A0);	/* PINCTRL40 */
	writel(0x0F0F0F0F, 0x5F8010A4);	/* PINCTRL41 */
	writel(0x0F0F0F0F, 0x5F8010A8);	/* PINCTRL42 */
	writel(0x0F0F0F0F, 0x5F8010AC);	/* PINCTRL43 */

	writel(0x3FC3FC79, 0x5F801A00);	/* PUPDCTRL0 */
	writel(0x0CC00000, 0x5F801A04);	/* PUPDCTRL1 */
	writel(0x00000000, 0x5F801A08);	/* PUPDCTRL2 */
	writel(0xF0000000, 0x5F801A0C);	/* PUPDCTRL3 */
	writel(0x8200001E, 0x5F801A10);	/* PUPDCTRL4 */
	writel(0x00000002, 0x5F801A14);	/* PUPDCTRL5 */

	writel(0x00000003, 0x5F801C04);	/* OECTRL */

	writel(0xFFFFFFFF, 0x5F801D00);	/* IECTRL0 */
	writel(0xFCCFFFFF, 0x5F801D04);	/* IECTRL1 */
	writel(0x000007FF, 0x5F801D08);	/* IECTRL2 */
	writel(0xF0000000, 0x5F801D0C);	/* IECTRL3 */
	writel(0xAF2F9F9E, 0x5F801D10);	/* IECTRL4 */
	writel(0x00001F02, 0x5F801D14);	/* IECTRL5 */

	writel(0x00000008, 0x55000008);	/* P0IOD */
	writel(0x000000E0, 0x5500000C);	/* P0DIR */
	writel(0x00000000, 0x550000D8);	/* P24IOD */
	writel(0x000000FE, 0x550000DC);	/* P24DIR */

	writel(0x00002431, 0x55000090);	/* IRQCTL */
	writel(0x00000000, 0x55000094);	/* IRQSEL */

#if (defined(CONFIG_AVE_RMII) || !defined(CONFIG_RGMII))
	writel(0x00000001, 0x5F800540);	/* ETPINMODE */
#else
	writel(0x00000000, 0x5F800540);	/* ETPINMODE */
#endif

	/*
	 * clocks must be enabled before deasserting resets
	 * (commit 2e5e1cc80a488bcb2e3482341c06d01578aa1246 of LD10 diag)
	 */
	for (i = 0; i < ARRAY_SIZE(clock_data); i++) {
		writel(clock_data[i].data, clock_data[i].address);
		udelay(10);
	}

	for (i = 0; i < ARRAY_SIZE(reset_data); i++) {
		writel(reset_data[i].data, reset_data[i].address);
		udelay(10);
	}

	/* Comment format:    PAD Name -> Function Name */

#if defined(CONFIG_NAND_UNIPHIER)
	sg_set_pinsel(3, 0, 8, 4);	/* XNFWP   -> XNFWP */
	sg_set_pinsel(4, 0, 8, 4);	/* XNFCE0  -> XNFCE0 */
	sg_set_pinsel(5, 0, 8, 4);	/* NFRYBY0 -> NFRYBY0 */
	sg_set_pinsel(6, 0, 8, 4);	/* XNFRE   -> XNFRE */
	sg_set_pinsel(7, 0, 8, 4);	/* XNFWE   -> XNFWE */
	sg_set_pinsel(8, 0, 8, 4);	/* NFALE   -> NFALE */
	sg_set_pinsel(9, 0, 8, 4);	/* NFCLE   -> NFCLE */
	sg_set_pinsel(10, 0, 8, 4);	/* NFD0    -> NFD0 */
	sg_set_pinsel(11, 0, 8, 4);	/* NFD1    -> NFD1 */
	sg_set_pinsel(12, 0, 8, 4);	/* NFD2    -> NFD2 */
	sg_set_pinsel(13, 0, 8, 4);	/* NFD3    -> NFD3 */
	sg_set_pinsel(14, 0, 8, 4);	/* NFD4    -> NFD4 */
	sg_set_pinsel(15, 0, 8, 4);	/* NFD5    -> NFD5 */
	sg_set_pinsel(16, 0, 8, 4);	/* NFD6    -> NFD6 */
	sg_set_pinsel(17, 0, 8, 4);	/* NFD7    -> NFD7 */
	sg_set_iectrl_range(3, 17);

	/* reset NANDC when bootmode_sel is nand only. */
	if (((readl(SG_PINMON0) >> 1) & 0x1f) <= 0x17) {
		u32 tmp;

		tmp = readl(SC_RSTCTRL4);
		tmp &= ~(0x1 << 0);
		writel(tmp, SC_RSTCTRL4);
		udelay(10);
		tmp |= (0x1 << 0);
		writel(tmp, SC_RSTCTRL4);
	}
#endif

#if defined(CONFIG_AVE_RMII) || !defined(CONFIG_RGMII)
	/* Ether */
	sg_set_pinsel(30, 0, 8, 4);	/* MDC         -> MDC */
	sg_set_pinsel(31, 0, 8, 4);	/* MDIO        -> MDIO */
	sg_set_pinsel(32, 0, 8, 4);	/* MDIO_INL    -> MDIO_INTL */
	sg_set_pinsel(33, 15, 8, 4);	/* PHYRSTL     -> PORT36 */
	sg_set_pinsel(34, 1, 8, 4);	/* RGMII_RXCLK -> RMII_REFCLK */
	sg_set_pinsel(35, 1, 8, 4);	/* RGMII_RXD0  -> RMII_RXD0 */
	sg_set_pinsel(36, 1, 8, 4);	/* RGMII_RXD1  -> RMII_RXD1 */
	sg_set_pinsel(37, 1, 8, 4);	/* RGMII_RXD2  -> RMII_RXER */
	sg_set_pinsel(38, 15, 8, 4);	/* RGMII_RXD3  -> PORT43 */
	sg_set_pinsel(39, 1, 8, 4);	/* RGMII_RXCTL -> RMII_CRS_DV */
	sg_set_pinsel(40, 15, 8, 4);	/* RGMII_TXCLK -> PORT45 */
	sg_set_pinsel(41, 1, 8, 4);	/* RGMII_TXD0  -> RMII_TXD0 */
	sg_set_pinsel(42, 1, 8, 4);	/* RGMII_TXD1  -> RMII_TXD1 */
	sg_set_pinsel(43, 15, 8, 4);	/* RGMII_TXD2  -> PORT50 */
	sg_set_pinsel(44, 15, 8, 4);	/* RGMII_TXD3  -> PORT51 */
	sg_set_pinsel(45, 1, 8, 4);	/* RGMII_TXCTL -> RMII_TXEN */
	sg_set_iectrl_range(30, 37);
	sg_set_iectrl(39);
	sg_set_iectrl(41);
	sg_set_iectrl(42);
	sg_set_iectrl(45);

	{
		u32 tmp;

		tmp = readl(0x55000024);
		tmp &= ~(0x1 << 6);
		writel(tmp, 0x55000024);	/* P3DIR6: OUT */

		tmp = readl(0x55000020);
		tmp |= (0x1 << 6);
		writel(tmp, 0x55000020);	/* P3IOD6: 1 */
	}
#else
	sg_set_pinsel(30, 0, 8, 4);	/* MDC         -> MDC */
	sg_set_pinsel(31, 0, 8, 4);	/* MDIO        -> MDIO */
	sg_set_pinsel(32, 0, 8, 4);	/* MDIO_INL    -> MDIO_INTL */
	sg_set_pinsel(33, 0, 8, 4);	/* PHYRSTL     -> PHYRSTL */
	sg_set_pinsel(34, 0, 8, 4);	/* RGMII_RXCLK -> RGMII_RXCLK */
	sg_set_pinsel(35, 0, 8, 4);	/* RGMII_RXD0  -> RGMII_RXD0 */
	sg_set_pinsel(36, 0, 8, 4);	/* RGMII_RXD1  -> RGMII_RXD1 */
	sg_set_pinsel(37, 0, 8, 4);	/* RGMII_RXD2  -> RGMII_RXD2 */
	sg_set_pinsel(38, 0, 8, 4);	/* RGMII_RXD3  -> RGMII_RXD3 */
	sg_set_pinsel(39, 0, 8, 4);	/* RGMII_RXCTL -> RGMII_RXCTL */
	sg_set_pinsel(40, 0, 8, 4);	/* RGMII_TXCLK -> RGMII_TXCLK */
	sg_set_pinsel(41, 0, 8, 4);	/* RGMII_TXD0  -> RGMII_TXD0 */
	sg_set_pinsel(42, 0, 8, 4);	/* RGMII_TXD1  -> RGMII_TXD1 */
	sg_set_pinsel(43, 0, 8, 4);	/* RGMII_TXD2  -> RGMII_TXD2 */
	sg_set_pinsel(44, 0, 8, 4);	/* RGMII_TXD3  -> RGMII_TXD3 */
	sg_set_pinsel(45, 0, 8, 4);	/* RGMII_TXCTL -> RGMII_TXCTL */
	sg_set_iectrl_range(30, 45);

	sg_set_drv3ctrl(28, 3);		/* RGMII_TXCLK */
	sg_set_drv3ctrl(29, 5);		/* RGMII_TXD0 */
	sg_set_drv3ctrl(30, 5);		/* RGMII_TXD1 */
	sg_set_drv3ctrl(31, 5);		/* RGMII_TXD2 */
	sg_set_drv3ctrl(32, 5);		/* RGMII_TXD3 */
	sg_set_drv3ctrl(33, 5);		/* RGMII_TXCTL */
#endif

	sg_set_pinsel(149, 14, 8, 4);	/* XIRQ0    -> XIRQ0 */
	sg_set_iectrl(149);
	sg_set_pinsel(153, 14, 8, 4);	/* XIRQ4    -> XIRQ4 */
	sg_set_iectrl(153);
}
