/*
 * sni_ave.c
 * Copyright (c) 2016-2017 Socionext inc.
 *
 * Author: Kunihiko Hayashi <hayashi.kunihiko@socionext.com>
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

#include <asm/io.h>
#include <common.h>
#include <dm.h>
#include <fdt_support.h>
#include <linux/mii.h>
#include <malloc.h>
#include <miiphy.h>
#include <net.h>

#include "sni_ave.h"

enum desc_id {
	AVE_DESCID_TX = 0,
	AVE_DESCID_RX,
};

enum desc_bit {
	AVE_DESC_32BIT = 0,
	AVE_DESC_64BIT,
};

struct ave_private {
	struct mii_dev *bus;
	struct phy_device *phydev;

	phys_addr_t iobase;

	int max_speed;
	int phy_mode;

	enum desc_bit is_desc_64bit;	/* 1 if description bitwidth is 64 */

	int rx_pos;	/* Current RX Descriptor pos. */
	int rx_siz;	/* RX buf size */
	int rx_off;	/* RX offset */
	int tx_num;	/* number of txed packets */

	uchar  tx_adj_packetbuf[PKTSIZE_ALIGN + PKTALIGN];
	uchar *tx_adj_buf;
};

static u32 ave_r32(struct ave_private *priv, u32 offset)
{
	return readl(priv->iobase + offset);
}

static void ave_w32(struct ave_private *priv, u32 offset, u32 value)
{
	writel(value, priv->iobase + offset);
}

static u32 ave_rdesc(struct ave_private *priv, enum desc_id id,
		     int entry, int offset)
{
	int desc_size;
	u32 addr;

	if (priv->is_desc_64bit) {
		desc_size = 12;
		addr = (id == AVE_DESCID_TX) ? AVE_TXDM_64 : AVE_RXDM_64;
	} else {
		desc_size = 8;
		addr = (id == AVE_DESCID_TX) ? AVE_TXDM_32 : AVE_RXDM_32;
	}

	addr += entry * desc_size + offset * 4;

	return ave_r32(priv, addr);
}

static void ave_wdesc(struct ave_private *priv, enum desc_id id,
		      int entry, int offset, u32 val)
{
	int desc_size;
	u32 addr;

	if (priv->is_desc_64bit) {
		desc_size = 12;
		addr = (id == AVE_DESCID_TX) ? AVE_TXDM_64 : AVE_RXDM_64;
	} else {
		desc_size = 8;
		addr = (id == AVE_DESCID_TX) ? AVE_TXDM_32 : AVE_RXDM_32;
	}

	addr += entry * desc_size + offset * 4;

	ave_w32(priv, addr, val);
}

static int ave_mdio_busywait(struct ave_private *priv)
{
	int ret = 1, loop = 1000;
	u32 mdiosr;

	/* wait until completion */
	while (1) {
		mdiosr = ave_r32(priv, AVE_MDIOSR);
		if (!(mdiosr & AVE_MDIOSR_STS))
			break;

		udelay(10); /* overhead=0.5usec. */
		if (!loop--) {
			printf("failed to read from MDIO (status:0x%08x)\n",
			       le32_to_cpu(mdiosr));
			ret = 0;
			break;
		}
	}

	return ret;
}

static int _ave_mdio_read(struct ave_private *priv, int phyid, int reg)
{
	u32 mdioctl;

	/* write address */
	ave_w32(priv, AVE_MDIOAR, (((phyid & 31) << 8) | (reg & 31)));

	/* read request */
	mdioctl = ave_r32(priv, AVE_MDIOCTR);
	ave_w32(priv, AVE_MDIOCTR, (mdioctl | AVE_MDIOCTR_RREQ));

	if (!ave_mdio_busywait(priv)) {
		printf("phy-%d reg-%x read failed\n\n",
		       phyid, reg);
		return 0xffff;
	}

	return ave_r32(priv, AVE_MDIORDR) & 0xffff;
}

static int _ave_mdio_write(struct ave_private *priv,
			   int phyid, int reg, u16 value)
{
	u32 mdioctl;

	/* write address */
	ave_w32(priv, AVE_MDIOAR, (((phyid & 31) << 8) | (reg & 31)));

	/* write data */
	ave_w32(priv, AVE_MDIOWDR, value);

	/* write request */
	mdioctl = ave_r32(priv, AVE_MDIOCTR);
	ave_w32(priv, AVE_MDIOCTR, (mdioctl | AVE_MDIOCTR_WREQ));

	if (!ave_mdio_busywait(priv)) {
		printf("phy-%d reg-%x write failed\n\n",
		       phyid, reg);
		return -1;
	}

	return 0;
}

static int ave_mdio_read(struct mii_dev *bus,
			 int addr, int devad, int reg)
{
	struct ave_private *priv = bus->priv;

	return _ave_mdio_read(priv, addr, reg);
}

static int ave_mdio_write(struct mii_dev *bus,
			  int addr, int devad, int reg, u16 val)
{
	struct ave_private *priv = bus->priv;

	return _ave_mdio_write(priv, addr, reg, val);
}

static int ave_adjust_link(struct ave_private *priv)
{
	struct phy_device *phydev = priv->phydev;
	u32 val, txcr, rxcr, rxcr_org;

	/* set RGMII speed */
	val = ave_r32(priv, AVE_TXCR);
	val &= ~(AVE_TXCR_TXSPD_100 | AVE_TXCR_TXSPD_1G);
	if (phydev->speed == 1000)
		val |= AVE_TXCR_TXSPD_1G;
	else if (phydev->speed == 100)
		val |= AVE_TXCR_TXSPD_100;
	ave_w32(priv, AVE_TXCR, val);

	/* set RMII speed */
	if (priv->phy_mode != PHY_INTERFACE_MODE_RGMII) {
		val = ave_r32(priv, AVE_LINKSEL);
		if (phydev->speed == 100)
			val |= AVE_LINKSEL_100M;
		else
			val &= ~AVE_LINKSEL_100M;
		ave_w32(priv, AVE_LINKSEL, val);
	}

	/* check current RXCR/TXCR */
	rxcr = ave_r32(priv, AVE_RXCR);
	txcr = ave_r32(priv, AVE_TXCR);
	rxcr_org = rxcr;

	if (phydev->duplex)
		rxcr |= AVE_RXCR_FDUPEN;
	else
		rxcr &= ~AVE_RXCR_FDUPEN;

	if (phydev->pause) {
		rxcr |= AVE_RXCR_FLOCTR;
		txcr |= AVE_TXCR_FLOCTR;
	} else {
		rxcr &= ~AVE_RXCR_FLOCTR;
		txcr &= ~AVE_TXCR_FLOCTR;
	}

	if (rxcr_org != rxcr) {
		/* disable Rx mac */
		ave_w32(priv, AVE_RXCR, rxcr & ~AVE_RXCR_RXEN);
		/* change and enable TX/Rx mac */
		ave_w32(priv, AVE_TXCR, txcr);
		ave_w32(priv, AVE_RXCR, rxcr);
	}

	printf("%s: phy:%s speed:%d\n",
	       priv->phydev->dev->name, phydev->drv->name, phydev->speed);

	return phydev->link;
}

static int ave_mdio_init(const char *name, struct ave_private *priv)
{
	struct mii_dev *bus = mdio_alloc();

	if (!bus) {
		printf("Failed to allocate MDIO bus\n");
		return -ENOMEM;
	}
	bus->read = ave_mdio_read;
	bus->write = ave_mdio_write;
	snprintf(bus->name, sizeof(bus->name), "%s", name);
	bus->priv = priv;

	return mdio_register(bus);
}

static int ave_phy_init(struct ave_private *priv, void *dev)
{
	struct phy_device *phydev;
	int mask = 0xffffffff, ret;

#if defined(CONFIG_PHY_ADDR)
	mask = 1 << CONFIG_PHY_ADDR;
#endif

	phydev = phy_find_by_mask(priv->bus, mask, priv->phy_mode);
	if (!phydev)
		return -ENODEV;

	phy_connect_dev(phydev, dev);

	phydev->supported &= PHY_GBIT_FEATURES;
	if (priv->max_speed) {
		ret = phy_set_supported(phydev, priv->max_speed);
		if (ret)
			return ret;
	}
	phydev->advertising = phydev->supported;

	priv->phydev = phydev;
	phy_config(phydev);

	return 0;
}

static void _ave_halt(struct ave_private *priv)
{
	u32 val;

	/* stop Tx Rx and GRST (w/o PHY RST) */
	val = ave_r32(priv, AVE_GRR);
	if (!val) {	/* when GRR all cleared */
		val = ave_r32(priv, AVE_RXCR);
		val &= ~AVE_RXCR_RXEN;
		ave_w32(priv, AVE_RXCR, val);
		ave_w32(priv, AVE_DESCC, AVE_DESCC_RXDSTP);
		do {
			val = ave_r32(priv, AVE_DESCC);
		} while (!(val & AVE_DESCC_RXDSTPST));
		ave_w32(priv, AVE_GRR, AVE_GRR_GRST);
	}
	mdelay(AVE_GRST_DELAY_MSEC);
}

static void _ave_reset(struct ave_private *priv)
{
	u32 val;

	/* RMII register reset */
	val = ave_r32(priv, AVE_RSTCTL);
	val &= (~AVE_RSTCTRL_RMIIRST);
	ave_w32(priv, AVE_RSTCTL, val);

	ave_w32(priv, AVE_GRR, AVE_GRR_CLR);
	mdelay(AVE_GRST_DELAY_MSEC);

	/* RMII register reset clear */
	val = ave_r32(priv, AVE_RSTCTL);
	val |= AVE_RSTCTRL_RMIIRST;
	ave_w32(priv, AVE_RSTCTL, val);
}

static int _ave_init(struct ave_private *priv)
{
	u32 val;
	uintptr_t val2;
	int i;

	_ave_reset(priv);

	priv->rx_pos = 0;
	priv->rx_off = 2; /* RX data has 2byte offsets */
	priv->tx_num = 0;
	priv->tx_adj_buf =
	    (uchar *)roundup((uintptr_t)&priv->tx_adj_packetbuf[0], PKTALIGN);
	priv->rx_siz = (PKTSIZE_ALIGN - priv->rx_off);

	val = 0;
	if (priv->phy_mode != PHY_INTERFACE_MODE_RGMII)
		val = AVE_CFGR_MII;
	ave_w32(priv, AVE_CFGR, val);

	/*
	 * set num descriptors
	 * use one TX descriptor and PKTBUFSRX of RX descriptors
	 */
	/* set a TX descriptor */
	ave_w32(priv, AVE_TXDC, AVE_TXDC_SIZE(priv, 1));
	ave_wdesc(priv, AVE_DESCID_TX, 0, AVE_TXDM_CMDSTS, 0);

	/* set RX descriptors */
	ave_w32(priv, AVE_RXDC, AVE_RXDC_SIZE(priv, PKTBUFSRX));
	for (i = 0; i < PKTBUFSRX; i++) {
		val2 = (uintptr_t)net_rx_packets[i];
		ave_cache_flush(val2, priv->rx_siz + priv->rx_off);

		ave_wdesc(priv, AVE_DESCID_RX, i,
			  AVE_RXDM_BUFPTR, (u32)(val2 & 0xFFFFFFFFULL));
		if (priv->is_desc_64bit)
			ave_wdesc(priv, AVE_DESCID_RX, i,
				  AVE_RXDM_BUFPTR+1, (u32)((u64)val2 >> 32));
		ave_wdesc(priv, AVE_DESCID_RX, i,
			  AVE_RXDM_CMDSTS, priv->rx_siz);
	}

	/* clear interrupt, set interrupt mask */
	ave_w32(priv, AVE_GISR, AVE_GISR_CLR);
	ave_w32(priv, AVE_GIMR, AVE_GIMR_CLR);

	/* set RXEN, Full Duplex, MTU value */
	val = AVE_RXCR_RXEN | AVE_RXCR_FDUPEN | AVE_RXCR_FLOCTR | AVE_RXCR_MTU;
	ave_w32(priv, AVE_RXCR, val);

	/* enable TX/RX descriptor */
	ave_w32(priv, AVE_DESCC, AVE_DESCC_RDE | AVE_DESCC_TDE);

	return 0;
}

static int _ave_write_hwaddr(struct ave_private *priv, u8 *mac)
{
	u32 val;

	val = le32_to_cpu(((u32 *)mac)[0]);
	ave_w32(priv, AVE_RXMAC1R, val);
	val = (u32)le16_to_cpu(((u16 *)mac)[2]);
	ave_w32(priv, AVE_RXMAC2R, val);

	return 0;
}

static int _ave_send(struct ave_private *priv, void *packet, int length)
{
	u32 val;
	uintptr_t val2;
	uchar *ptr = packet;

	/* adjust alignment for descriptor */
	if ((uintptr_t)ptr & 0x3) {
		memcpy((void *)priv->tx_adj_buf, (const void *)ptr, length);
		ptr = priv->tx_adj_buf;
	}

	/* padding for minimum length */
	if (length < AVE_MIN_XMITSIZE) {
		memset((void *)(ptr + length), 0, AVE_MIN_XMITSIZE - length);
		length = AVE_MIN_XMITSIZE;
	}

	/* check ownership and wait for previous xmit done */
	do {
		val = ave_rdesc(priv, AVE_DESCID_TX, 0, AVE_TXDM_CMDSTS);
	} while (val & AVE_TXDM_OWN);

	ave_cache_flush((uintptr_t)ptr, length);

	/* set addr */
	val2 = (uintptr_t)ptr;
	ave_wdesc(priv, AVE_DESCID_TX, 0,
		  AVE_TXDM_BUFPTR, (u32)(val2 & 0xFFFFFFFFULL));
	if (priv->is_desc_64bit)
		ave_wdesc(priv, AVE_DESCID_TX, 0,
			  AVE_TXDM_BUFPTR + 1, (u32)((u64)val2 >> 32));

	/* set size and kick xmit */
	val = AVE_TXDM_OWN | length | AVE_TX_CMDSTS_1ST | AVE_TX_CMDSTS_LAST;
	ave_wdesc(priv, AVE_DESCID_TX, 0, AVE_TXDM_CMDSTS, val);
	priv->tx_num++;

	/* wait for xmit */
	do {
		val = ave_rdesc(priv, AVE_DESCID_TX, 0, AVE_TXDM_CMDSTS);
	} while (val & AVE_TXDM_OWN);

	if (!(val & AVE_TXDM_OK))
		printf("%s: bad send status=0x%08x\n",
		       priv->phydev->dev->name, le32_to_cpu(val));

	return 0;
}

static int _ave_recv(struct ave_private *priv, uchar **packetp)
{
	uchar *ptr;
	int length = 0;
	u32 cmdsts;

	while (1) {
		cmdsts = ave_rdesc(priv, AVE_DESCID_RX,
				   priv->rx_pos, AVE_RXDM_CMDSTS);
		if (!(cmdsts & AVE_RXDM_OWN))
			/* hardware ownership, no recieved packets */
			return -EAGAIN;

		ptr = net_rx_packets[priv->rx_pos] + priv->rx_off;
		if (cmdsts & AVE_RXDM_OK) {
			length = cmdsts & AVE_RXDM_SIZE_MASK;
			/* invalidate after DMA is done */
			ave_cache_invalidate((uintptr_t)ptr, length);
			*packetp = ptr;
			break;
		}
		printf("%s: bad packet[%d] status=0x%08x ptr=%p\n",
		       priv->phydev->dev->name, priv->rx_pos,
		       le32_to_cpu(cmdsts), ptr);
	}

	return length;
}

static int _ave_free_packet(struct ave_private *priv)
{
	/* flush before DMA starts */
	ave_cache_flush((uintptr_t)net_rx_packets[priv->rx_pos],
			priv->rx_siz + priv->rx_off);

	/* set hardware ownership, kick receive */
	ave_wdesc(priv, AVE_DESCID_RX,
		  priv->rx_pos, AVE_RXDM_CMDSTS, priv->rx_siz);

	/* set to next recieving position */
	if (++priv->rx_pos >= PKTBUFSRX)
		priv->rx_pos = 0;

	return 0;
}

#if defined(CONFIG_DM_ETH)

static int ave_start(struct udevice *dev)
{
	struct ave_private *priv = dev_get_priv(dev);
	int ret;

	ret = _ave_init(priv);
	if (ret)
		return ret;

	phy_startup(priv->phydev);
	ave_adjust_link(priv);

	return 0;
}

static void ave_stop(struct udevice *dev)
{
	struct ave_private *priv = dev_get_priv(dev);

	_ave_halt(priv);
	phy_shutdown(priv->phydev);
}

static int ave_send(struct udevice *dev, void *packet, int length)
{
	return _ave_send(dev_get_priv(dev), packet, length);
}

static int ave_recv(struct udevice *dev, int flags, uchar **packetp)
{
	return _ave_recv(dev_get_priv(dev), packetp);
}

static int ave_free_packet(struct udevice *dev, uchar *packet, int length)
{
	return _ave_free_packet(dev_get_priv(dev));
}

static int ave_write_hwaddr(struct udevice *dev)
{
	struct ave_private *priv = dev_get_priv(dev);
	struct eth_pdata *pdata = dev_get_platdata(dev);
	u8 *mac = pdata->enetaddr;

	return _ave_write_hwaddr(priv, mac);
}

static int ave_ofdata_to_platdata(struct udevice *dev)
{
	struct ave_private *priv = dev_get_priv(dev);
	struct eth_pdata *pdata = dev_get_platdata(dev);
	const char *phy_mode;
	const u32 *valp;
	u32 val;

	pdata->iobase = devfdt_get_addr(dev);
	pdata->phy_interface = -1;
	phy_mode = fdt_getprop(gd->fdt_blob, dev_of_offset(dev), "phy-mode",
			       NULL);
	if (phy_mode)
		pdata->phy_interface = phy_get_interface_by_name(phy_mode);
	if (pdata->phy_interface == -1) {
		printf("Invalid PHY interface '%s'\n", phy_mode);
		return -EINVAL;
	}

	pdata->max_speed = 0;
	valp = fdt_getprop(gd->fdt_blob, dev_of_offset(dev), "max_speed",
			   NULL);
	if (valp)
		pdata->max_speed = fdt32_to_cpu(*valp);

	return 0;
}

static int ave_probe(struct udevice *dev)
{
	struct eth_pdata *pdata = dev_get_platdata(dev);
	struct ave_private *priv = dev_get_priv(dev);

	priv->iobase = pdata->iobase;
	priv->phy_mode = pdata->phy_interface;
	priv->max_speed = pdata->max_speed;
	priv->is_desc_64bit = dev_get_driver_data(dev);

	_ave_reset(priv);
	ave_mdio_init(dev->name, priv);
	priv->bus = miiphy_get_dev_by_name(dev->name);

	return ave_phy_init(priv, dev);
}

static int ave_remove(struct udevice *dev)
{
	struct ave_private *priv = dev_get_priv(dev);

	free(priv->phydev);
	mdio_unregister(priv->bus);
	mdio_free(priv->bus);

	return 0;
}

static const struct eth_ops ave_ops = {
	.start        = ave_start,
	.stop         = ave_stop,
	.send         = ave_send,
	.recv         = ave_recv,
	.free_pkt     = ave_free_packet,
	.write_hwaddr = ave_write_hwaddr,
};

static const struct udevice_id ave_ids[] = {
	{ .compatible = "socionext,uniphier-ld20-ave4", .data = AVE_DESC_64BIT, },
	{ .compatible = "socionext,uniphier-ld11-ave4", .data = AVE_DESC_32BIT, },
	{ .compatible = "socionext,uniphier-pxs2-ave4", .data = AVE_DESC_32BIT, },
	{ .compatible = "socionext,uniphier-pxs3-ave4", .data = AVE_DESC_32BIT, },
	{ /* Sentinel */ }
};

U_BOOT_DRIVER(ave) = {
	.name     = "ave",
	.id       = UCLASS_ETH,
	.of_match = ave_ids,
	.probe	  = ave_probe,
	.remove	  = ave_remove,
	.ofdata_to_platdata = ave_ofdata_to_platdata,
	.ops	  = &ave_ops,
	.priv_auto_alloc_size = sizeof(struct ave_private),
	.platdata_auto_alloc_size = sizeof(struct eth_pdata),
};

#else /* CONFIG_DM_ETH */

static int ave_init(struct eth_device *dev, bd_t *bd)
{
	struct ave_private *priv = (struct ave_private *)dev->priv;
	int ret;

	ret = _ave_init(priv);
	if (ret)
		return ret;

	phy_startup(priv->phydev);
	ave_adjust_link(priv);

	return 0;
}

static void ave_halt(struct eth_device *dev)
{
	struct ave_private *priv = (struct ave_private *)dev->priv;

	_ave_halt(priv);

	phy_shutdown(priv->phydev);
}

static int ave_send(struct eth_device *dev, void *packet, int length)
{
	return _ave_send(dev->priv, packet, length);
}

static int ave_recv(struct eth_device *dev)
{
	uchar *packet;
	int length;

	for (;;) {
		length = _ave_recv(dev->priv, &packet);
		if (length >= 0) {
			net_process_received_packet(packet, length);
			_ave_free_packet(dev->priv);
		} else if (length < 0) {
			return length;
		}
	}
}

static int ave_write_hwaddr(struct eth_device *dev)
{
	return _ave_write_hwaddr(dev->priv, dev->enetaddr);
}

int ave_initialize(u8 dev_num, int base_addr)
{
	struct eth_device *dev;
	struct ave_private *priv;
	u32 val;

	dev = malloc(sizeof(*dev));
	if (!dev)
		return 0;
	priv = malloc(sizeof(*priv));
	if (!priv) {
		free(dev);
		return 0;
	}

	memset(dev, 0, sizeof(*dev));
	dev->priv = priv;
	dev->iobase = base_addr;

	priv->iobase = base_addr;
	priv->phydev = NULL;
	priv->bus = NULL;

#if defined(CONFIG_SNI_AVE_DESC_64BIT)
	priv->is_desc_64bit = 1; /* 1 if description bitwidth is 64 */
#else
	priv->is_desc_64bit = 0; /* 1 if description bitwidth is 64 */
#endif

#if defined(CONFIG_RGMII)
	priv->phy_mode = PHY_INTERFACE_MODE_RGMII;
	priv->max_speed = 1000;
#else
	priv->phy_mode = PHY_INTERFACE_MODE_RMII;
	priv->max_speed = 100;
#endif

	/* check ID */
	val = ave_r32(priv, AVE_IDR);
	if (val != AVE_IDR_VALUE) {
		printf("%s: IDR fail(base=%#x)\n", __func__, base_addr);
		free(priv);
		free(dev);
		return -ENODEV;
	}

	dev->init = ave_init;
	dev->halt = ave_halt;
	dev->send = ave_send;
	dev->recv = ave_recv;
	dev->write_hwaddr = ave_write_hwaddr;
	sprintf(dev->name, "%s-%hu", DRIVERNAME, dev_num);

	eth_register(dev);

	_ave_reset(priv);
	ave_mdio_init(dev->name, priv);
	priv->bus = miiphy_get_dev_by_name(dev->name);

	return ave_phy_init(priv, dev);
}
#endif /* CONFIG_DM_ETH */
